from medgpt.datasets.utils import get_all_splits
from collections import defaultdict


def filter_by_count(dataset, min_count=5, min_count_global=100, min_length=5, num_proc=None):
    r''' Filters tokens of a dataset and leaves only the ones with frequencey >= min_count

    Args:
        dataset
        min_count:
            Intra patient count
        min_count_global:
            Whole dataset count
        min_length:
            Examples below will be removed, in other words patients with less than min_length concepts
        num_proc
    '''
    token_cnt = defaultdict(int)
    for _dataset in get_all_splits(dataset):
        for stream in _dataset['stream']:
            seen_in_stream = set()
            for sample in stream:
                token = sample['token']
                if token not in seen_in_stream:
                    token_cnt[token] += 1
                    seen_in_stream.add(token)

    if min_count_global is not None:
        # First we filter by global count, ie a concept has to have more than min_count_global appearances in the whole dataset
        dataset = dataset.map(function=lambda example: {'stream': [sample for sample in example['stream'] if token_cnt[sample['token']] >= min_count_global]},
                          load_from_cache_file=False, num_proc=num_proc)

    if min_count is not None:
        # Next, filter by intra-patient count
        dataset = dataset.map(function=lambda example: {'stream': [sample for sample in example['stream'] if sample['cnt'] >= min_count]},
                          load_from_cache_file=False, num_proc=num_proc)

    # Remove short streams
    if min_length > 0:
        dataset = dataset.filter(function=lambda example: len(example['stream']) >=  min_length, num_proc=num_proc, load_from_cache_file=False)

    return dataset



import numpy as np
import pandas as pd


def precision(predictions, label_ids, id2tkn, token_type2tokens, type_data, select_token_types={'T-11'}, prediction_scope='one', shifted_labels=False,
        predictions_are_scores=True, old_data=None, topk=1, start=0, time_range=None, time_data=None):
    r''' Calculate precision for next concept prediction.

    Args:
        predictions:
            Expected shape <batch_size> x <sequence_length> x <vocabulary_size>
        label_ids:
            Expected shape <batch_size> x <sequence_length>
        token_type2tokens:
            Map from a token type to all tokens belonging to it
        type_data:
            token types for each label/example
        select_token_types (Set[str], optional, defaults to `{'cui'}`:
            On what token types to calculate the Precision. Leave empoty to include all token types.
        prediction_scope:
            How much into the future should we look to accept something as correct:
                - `one` has to be the next concept
                - `age` until the next age token
                - `any` whenever
        shifted_labels:
            Are labels == input_ids, or shifted by one to the left
        predictions_are_scores:
            Are predictions scores for each label_id or really label_ids already
        old_data:
            If set it will load old values for tp/fp/positives/negatives and continue ontop of those
        topk:
            How many predicted labels to consider when calculating precision
        start:
            At what point to start - we will look only at the precision of concepts at positions after start

    Return (Dict[str, ]):
        precision:
            Precision
        tp:
            Number of True positives
        fp:
            Number of False positives
        positives:
            For each label ID a count of positive examples
        negatives
            For each label ID a count of negative examples
    '''
    if predictions_are_scores:
        if type(predictions) == list:
            outputs = [np.argsort(-1 * x, axis=1) for x in predictions]
        else:
            # TODO: do some validation here, as this will let any type of predictions through
            outputs = np.argsort(-1 * predictions, axis=2)
    else:
        outputs = predictions
    tp = {'all': 0, 'new': 0, 'old': 0}
    fp = {'all': 0, 'new': 0, 'old': 0}
    positives = {'all': {}, 'new': {}, 'old': {}}
    negatives = {'all': {}, 'new': {}, 'old': {}}
    numerical_errors = []

    # Are the requested token types or numerical and can a numerical error be calculated
    calculate_numerical_error = all([tkn_type in ['age', 'ttd'] for tkn_type in select_token_types])

    # If not shifted_labels label = prediction - 1
    label_position_shift = 0 if shifted_labels else 1
    # If labels are not shifted move the start by one
    start += 0 if shifted_labels else 1

    if old_data:
        tp = old_data['tp']
        fp = old_data['fp']
        positives = old_data['positives']
        negatives = old_data['negatives']
        numerical_errors = old_data['numerical_errors']

    def prediction_end_index(i, lbl, ind):
        r''' Used below to get the end index for different
        prediction scopes
        '''
        if prediction_scope == 'one':
            return i + 1
        elif prediction_scope == 'any':
            return len(lbl)
        elif prediction_scope == 'age':
            end = len(lbl) # Set end to last token in the labels array (for one example)
            _token_types = type_data[ind]
            for j in range(i, len(lbl)):
                type_label = _token_types[j] if j < len(_token_types) else 'unk'
                if type_label == 'age':
                    end = j
                    break
            return end
        elif prediction_scope == 'sep':
            end = len(lbl) # Set end to last token in the labels array (for one example)
            _token_types = type_data[ind]
            for j in range(i, len(lbl)):
                type_label = _token_types[j] if j < len(_token_types) else 'unk'
                if type_label == 'sep':
                    end = j
                    break
            return end
        elif prediction_scope == 'time_range':
            end = len(lbl) # Set end to last token in the labels array (for one example)
            token_time = time_data[ind]
            for j in range(i, len(lbl)):
                if j < len(token_time): # It can be that time is not available for padding tokens
                    if token_time[j] > (token_time[i] + time_range):
                        end = j
                        break
            return end

    for ind, lbl in enumerate(label_ids):
        _token_types = type_data[ind]
        if start < len(lbl):
            for i in range(start, len(lbl)):
                tkn_label = str(id2tkn.get(lbl[i], lbl[i]))
                type_label = _token_types[i] if i < len(_token_types) else 'unk'
                if type_label in select_token_types:
                    candidates = []
                    select_tokens = token_type2tokens[type_label]
                    if predictions_are_scores:
                        # We only get the type of canidate we know we need at this position
                        for k in range(len(outputs[ind][i-label_position_shift])):
                            if id2tkn[outputs[ind][i-label_position_shift][k]] in select_tokens:
                                candidates.append(outputs[ind][i-label_position_shift][k])

                            if len(candidates) == topk:
                                break
                    else:
                        candidates.append(outputs[ind][i-label_position_shift])

                    is_tp = False
                    is_new = False
                    end = prediction_end_index(i, lbl, ind)
                    for candidate in candidates:
                        # Is it a new concept or an existing one, this only makes sense when there
                        #is just one candidate
                        if candidate not in lbl[0:i]:
                            is_new = True
                        else:
                            is_new = False

                        tkn_candidate = str(id2tkn.get(candidate, candidate))
                        if tkn_candidate in select_tokens:
                            # If predictions are scores we can do topk, if not just do simple label match
                            if candidate in lbl[i:end]:
                                positives['all'][tkn_candidate] = positives['all'].get(tkn_candidate, 0) + 1
                                is_tp = True
                            else:
                                negatives['all'][tkn_candidate] = negatives['all'].get(tkn_candidate, 0) + 1
                        else:
                            negatives['all'][tkn_candidate] = negatives['all'].get(tkn_candidate, 0) + 1
                    if is_tp:
                        tp['all'] += 1

                        temporality = 'new' if is_new else 'old'
                        tp[temporality] += 1
                        positives[temporality][tkn_candidate] = positives[temporality].get(tkn_candidate, 0) + 1
                    else:
                        fp['all'] += 1

                        temporality = 'new' if is_new else 'old'
                        fp[temporality] += 1
                        negatives[temporality][tkn_candidate] = negatives[temporality].get(tkn_candidate, 0) + 1

                    if calculate_numerical_error:
                        # Both have to be of the right type, that is how candidates are setup
                        num_label = int(tkn_label)
                        num_pred = int(tkn_candidate)
                        numerical_error = abs(num_label - num_pred)
                        numerical_errors.append([num_label, num_pred, numerical_error])


    precision = {}
    for temporality in tp.keys():
        if tp[temporality] > 0:
            precision[temporality] = tp[temporality] / (tp[temporality] + fp[temporality])
        else:
            precision[temporality] = 0

    return {
            'precision': precision,
            'tp': tp,
            'fp': fp,
            'positives': positives,
            'negatives': negatives,
            'numerical_errors': numerical_errors,
            }


def metrics_data2df(metrics_data, cdb, main='positives', temporality='all'):
    d = metrics_data
    if main == 'positives':
        other = 'negatives'
    else:
        other = 'positives'

    out = sorted([(
        "{:.2f}".format(tp / (tp + d[other][temporality].get(cui, 0))),
        cdb.get_name(cui),
        cui,
        tp,
        d[other][temporality].get(cui, 0),
        cui) for cui, tp in sorted(d[main][temporality].items(), key=lambda x: x[1], reverse=True)],
        key=lambda x: x[0], reverse=True)

    out = pd.DataFrame(out, columns=['precision', 'name', 'cui', main, other, 'cui'])

    return out


class ComputePrecisionHF(object):
    r''' Used for computing precison when working with HF trainer
    '''

    def __init__(self, id2tkn, type_data, token_type2tokens, batch_size=1000, topk=1, return_all_metrics=False, time_range=None, time_data=None, **kwargs):
        self.id2tkn = id2tkn
        self.batch_size = batch_size
        self.kwargs = kwargs
        self.topk = topk
        self.return_all_metrics = return_all_metrics
        self.type_data = type_data
        self.token_type2tokens = token_type2tokens
        self.time_range = time_range
        self.time_data = time_data

    def __call__(self, p):
        # We will do this in batches, because it can be very memory demanding
        metrics_data = None
        start = 0
        while start < len(p.predictions):
            predictions = p.predictions[start:start+self.batch_size]
            label_ids = p.label_ids[start:start+self.batch_size]
            if self.time_data is not None:
                time_data_batch = self.time_data[start:start+self.batch_size]
            else:
                time_data_batch = None
            type_data_batch = self.type_data[start:start+self.batch_size]

            metrics_data = precision(predictions, label_ids=label_ids, token_type2tokens=self.token_type2tokens,
                                     id2tkn=self.id2tkn, type_data=type_data_batch, old_data=metrics_data,
                                     predictions_are_scores=True, topk=self.topk, time_range=self.time_range,
                                     time_data=time_data_batch, **self.kwargs)
            start += self.batch_size

        if self.return_all_metrics:
            return {
                'metrics_data': metrics_data, # Return all the metrics data too
            }
        else:
            return {
                'precision': metrics_data['precision']['all'], # Return only the overall precision
            }



import random
import torch

class SimpleMapTokenizer(object):
    r''' Not even really a tokenizer, will take a list of tokens and
    covert them to IDs

    Args:
        tkn2id
        pad_id
        max_len
    '''
    def __init__(self, tkn2id, pad_id, max_len=50, tkn2name=None):
        self.tkn2id = tkn2id
        self.pad_id = pad_id
        self.max_len = max_len
        self.tkn2name = tkn2name

        # Create id2tkn 
        self.id2tkn = {v:k for k,v in self.tkn2id.items()}


    def __call__(self, text, return_tensors=False):
        out = {'input_ids': [], 'attention_mask': []}

        out['input_ids'] = out['input_ids'] + [self.tkn2id[tkn] for tkn in text.split(" ")]
        if self.end_id is not None:
            out['input_ids'] = out['input_ids'] + [self.end_id]

        out['attention_mask'] = [1] * len(out['input_ids'])

        if return_tensors:
            out = {k:torch.tensor([v]) for k,v in out.items()}

        return out


    def decode(self, token_ids, get_names=True):
        tkns = self.convert_ids2tokens(token_ids, get_names=get_names)
        if type(tkns) != list:
            tkns = [tkns]
        return " ".join(tkns)


    def convert_ids2tokens(self, token_ids, get_names=True):
        if type(token_ids) == torch.Tensor:
            token_ids = token_ids.tolist()
        if type(token_ids) == list and type(token_ids[0]) == torch.Tensor:
            token_ids = [x.tolist() for x in token_ids]

        # Same as decode, but needed for compatibility with ecco
        out = []
        if type(token_ids) != list:
            out = [self.id2tkn[int(token_ids)]]
        else:
            # Convert tokens to IDs
            out = [self.id2tkn[int(id)] for id in token_ids]

        if get_names:
            _out = []
            for tkn in out:
                _out.append(self.tkn2name.get(tkn, tkn))
                #_out.append(" | ")
            out = _out

        return out



    def tokens_to_ids(self, tokens):
        r''' This will skip tokens if they are not in the tkn2id dict
        '''
        out = [self.tkn2id[tkn] for tkn in tokens if tkn in self.tkn2id]
        out = out[:self.max_len]

        return out


    def encode(self, examples):
        r''' Convert 'stream' in the examples from tokens to IDs, save as 'input_ids'. Use with HF datasets.map
        '''
        examples['input_ids'] = [self.tokens_to_ids(stream) for stream in examples['stream']]

        return examples



def get_parents_map(cuis, pt2ch, depth=3):
    r''' Get a map from a concept to all of its parents up to the `depth`, meaning parents of parents and so on.

    Args:
        pt2ch (`Dict`):
            map from parent concept to children (this is
            usually what we have when building a CDB).

        depth (`int`, optional defaults to 3):
            Get only parents, or parents of parents also, or ...
    '''

    # First convert pt2ch into ch2pt
    ch2pt = {}
    for pt in pt2ch:
        for ch in pt2ch[pt]:
            if ch in ch2pt:
                ch2pt[ch].add(pt)
            else:
                ch2pt[ch] = {pt}

    def get_parents(concept, ch2pt, depth):
        parents = set()
        parents.update(ch2pt.get(concept, []))
        if depth > 0:
            for pt in ch2pt.get(concept, []):
                parents.update(get_parents(pt, ch2pt, depth=depth-1))
        return parents

    ch2all_pt = {}
    for cui in cuis:
        ch2all_pt[cui] = get_parents(cui, ch2pt, depth=depth)

    return ch2all_pt



import logging
import numpy as np
from medcat.utils.matutils import unitvec
from datetime import datetime
import math
import datasets
import random
import copy


def get_all_splits(dataset):
    all_datasets = []
    if 'train' in dataset:
        all_datasets.append(dataset['train'])
    if 'test' in dataset:
        all_datasets.append(dataset['test'])
    if 'valid' in dataset:
        all_datasets.append(dataset['valid'])
    if isinstance(dataset, datasets.arrow_dataset.Dataset):
        # If we have only one, ie no train/test
        all_datasets.append(dataset)

    return all_datasets

def make_example(token, token_type='unk', cnt=10**6, time=None):
    return {'token': token, 'token_type': token_type, 'cnt': cnt, 'time': time}

def bucket_concepts(examples, bucket_size_seconds=365*24*60*60, separator='<SEP>'):
    r''' Will bucket concepts into specified bucket_size.

    Args:
        examples
    '''
    for i in range(len(examples['stream'])):
        stream = examples['stream'][i]

        new_stream = []
        _bucket = []
        _tokens = set()
        start_time = -1
        for ent in stream:
            if start_time == -1:
                start_time = ent['time']

            if ent['time'] - start_time >= bucket_size_seconds:
                # Add to stream
                new_stream.extend(_bucket)
                _bucket = []
                _tokens = set()
                start_time = ent['time']

                if separator is not None:
                    new_stream.append(make_example(token=separator, token_type='sep', cnt=10**6, time=new_stream[-1]['time'] + 1))

            if ent['token'] not in _tokens:
                _bucket.append(ent)
                _tokens.add(ent['token'])

        if _bucket:
            new_stream.extend(_bucket)

        examples['stream'][i] = new_stream
        new_stream = []

    return examples

def add_age(examples, pt2dob_timestamp, age_prefix='<AGE>', age_suffix=None, age_normalizer=365.25 * 24 * 60 * 60):
    for i in range(len(examples['stream'])):
        stream = examples['stream'][i]
        last_age_added = -1
        new_stream = []
        for ent in stream:
            if examples['patient_id'][i] in pt2dob_timestamp:
                age = int((ent['time'] - pt2dob_timestamp[examples['patient_id'][i]]) / age_normalizer)
                if last_age_added != age:
                    if age_prefix is not None:
                        new_stream.append(make_example(token=age_prefix, token_type='age_prefix', cnt=10**6, time=ent['time']))
                    new_stream.append(make_example(token=str(age), token_type='age', cnt=10**6, time=ent['time']))
                    last_age_added = age
                    if age_suffix is not None:
                        new_stream.append(make_example(token=age_suffix, token_type='age_suffx', cnt=10**6, time=ent['time']))

            new_stream.append(ent)

        examples['stream'][i] = new_stream
        new_stream = []

    return examples

def add_ttd(examples, pt2dod_timestamp, ttd_prefix='<TTD>', ttd_suffix=None, ttd_normalizer=365.25 * 24 * 60 * 60,
            max_ttd=10, ttd_prob=1, max_nttd=10, duplicate_streams=False):
    all_patient_id = []
    all_stream = []
    for i in range(len(examples['stream'])):
        stream = examples['stream'][i]
        last_ttd_added = -1
        new_stream = []
        new_streams = [new_stream]
        n_added_ttds = 0
        for ent in stream:
            if examples['patient_id'][i] in pt2dod_timestamp:
                if n_added_ttds < max_nttd:
                    if random.random() <=  ttd_prob:
                        ttd = int((pt2dod_timestamp[examples['patient_id'][i]] - ent['time']) / ttd_normalizer) + 1
                        if ttd <= max_ttd:
                            if last_ttd_added != ttd:
                                if duplicate_streams:
                                    # At this point we duplicate the first stream fron new_streams (it is the one without TTD always)
                                    new_stream = copy.deepcopy(new_streams[0])
                                    new_streams.append(new_stream)

                                if ttd_prefix is not None:
                                    new_stream.append(make_example(token=ttd_prefix, token_type='ttd_prefix', cnt=10**6, time=ent['time']))
                                new_stream.append(make_example(token=str(ttd), token_type='ttd', cnt=10**6, time=ent['time']))

                                last_ttd_added = ttd
                                if ttd_suffix is not None:
                                    new_stream.append(make_example(token=ttd_suffix, token_type='ttd_suffix', cnt=10**6, time=ent['time']))
                                n_added_ttds += 1

            # append the entity to each stream
            for new_stream in new_streams: new_stream.append(ent)

        if duplicate_streams and len(new_streams) > 1:
            # Remove the first example as it is the base one without time info
            del new_streams[0]

        for new_stream in new_streams:
            all_stream.append(new_stream)
            all_patient_id.append(examples['patient_id'][i])

    examples['patient_id'] = all_patient_id
    examples['stream'] = all_stream

    return examples

def split_stream(examples, max_seq_len=-1):
    if max_seq_len > 0:
        new_streams = []
        new_patient_ids = []
        for ind, stream in enumerate(examples['stream']):
            nparts = math.ceil(len(stream) / max_seq_len)
            for i in range(nparts):
                new_streams.append(stream[i*max_seq_len:(i+1)*max_seq_len])
                new_patient_ids.append(examples['patient_id'][ind])

        examples['stream'] = new_streams
        examples['patient_id'] = new_patient_ids

    return examples


def cleanup_stream(examples, keep_time=True, keep_type=True):
    r''' Leave only Tokens and remove the rest from `stream`

    Args:
        examples
        keep_time:
            If set another value will be added to examples that contains the `time` for each
            entity in stream.
        keep_type:
            Same as above
    '''
    if 'token' in examples['stream'][0][0]:
        if keep_time:
            examples['time'] = [[ent['time'] for ent in stream] for stream in examples['stream']]
        if keep_type:
            examples['token_type'] = [[ent['token_type'] for ent in stream] for stream in examples['stream']]
        examples['stream'] = [[ent['token'] for ent in stream] for stream in examples['stream']]

    return examples


def add_to_stream(examples, pt2tkn, last=False, prefix=None, unk_tkn='unk', token_type='unk'):
    r''' Add information to the patient stream based on patient_id.

    Args:
        examples
        pt2tkn
        last
        unk_tkn:
            What token will be added if the patient_id is not in pt2tkn
    '''

    for i in range(len(examples['stream'])):
        token = pt2tkn.get(examples['patient_id'][i], unk_tkn)
        to_append = [make_example(token=token, cnt=10**6, time=examples['stream'][i][0]['time'] - 1, token_type=token_type)]
        if prefix is not None:
            prefix_token = make_example(token=prefix, cnt=10**6, time=examples['stream'][i][0]['time'] - 1, token_type="prefix_" + token_type)
            to_append = [prefix_token] + to_append

        if last:
            # Append as last token
            examples['stream'][i] = examples['stream'][i] + to_append
        else:
            examples['stream'][i] = to_append + examples['stream'][i]

    return examples


def remove_parents_from_stream(examples, ch2parents, separator=None):
    for i in range(len(examples['stream'])):
        stream = examples['stream'][i]
        parents = set()
        new_stream = []

        for ent in stream:
            tkn = ent['token']

            if separator is not None and tkn == separator:
                # This means we are removing parents only inside of one bucket
                parents = set()

            if tkn in ch2parents:
                # Add only if not in parents
                if tkn not in parents:
                    new_stream.append(ent)
                # Update parents
                parents.update(ch2parents[tkn])
            else:
                new_stream.append(ent)

        examples['stream'][i] = new_stream

    return examples

def get_embeddings_for_tokens(dataset, cdb, context_type='medium', normalize=True, extra_tokens=['<PAD>']):
    r''' Given a stream of tokens get the embeddings from MedCAT and make the required maps.

    Args:
        dataset
        cdb
        context_type
        normalize:
            If True the embedding vectors will be normalized
        tkn2type:
            Dictionary mapping from token to type
    Returns:
        embeddings
        tkn2id
        id2tkn
        id2type
        id2type_detailed
    '''
    embeddings = []
    tkn2id = {}
    id2tkn = {}

    datasets = get_all_splits(dataset)
    for _dataset in datasets:
        for stream in _dataset['stream']:
            for tkn in stream:
                tkn = str(tkn)
                if tkn not in tkn2id:
                    if tkn in cdb.cui2context_vectors and context_type in cdb.cui2context_vectors[tkn]:
                        vec = cdb.cui2context_vectors[tkn][context_type]
                    else:
                        # Token vector is randomly assigned
                        vec = np.random.rand(300)

                    id2tkn[len(embeddings)] = tkn
                    tkn2id[tkn] = len(embeddings)

                    vec = unitvec(vec) if normalize else vec
                    embeddings.append(vec)

        # Add named tokens
    for tkn in extra_tokens:
        id2tkn[len(embeddings)] = tkn
        tkn2id[tkn] = len(embeddings)
        if tkn != '<PAD>':
            embeddings.append(np.random.rand(len(embeddings[0])))
        else:
            embeddings.append(np.zeros(len(embeddings[0])))

    return embeddings, tkn2id, id2tkn


def stream_to_separate_examples(examples):
    r''' Convert a stream to separate examples that can be used to train
    a next concept predictor unable to handle sequences (e.g. random forset). Use with HF datasets map function.

    '''
    out = {}
    out['input_ids'] = [input_ids[0:i+1] for input_ids in examples['input_ids'] for i in range(len(input_ids) - 1)]
    out['labels'] = [input_ids[i+1] for input_ids in examples['input_ids'] for i in range(len(input_ids) - 1)]
    out['labels_all'] = [input_ids[i+1:] for input_ids in examples['input_ids'] for i in range(len(input_ids) - 1)]
    out['patient_id'] = [patient_id for ind, patient_id in enumerate(examples['patient_id']) for _ in range(len(examples['input_ids'][ind]) - 1)]

    return out



from collections import defaultdict


def docs2stream(docs, doc2pt, doc2time=None, meta_requirements={}, entity_type_column='tuis'):
    r''' Convert the `docs` output of medcat multiprocessing
    to a stream of concepts for each patient.

    Args:
        docs
        doc2pt
        doc2time
        meta_requirements:
            Values for meta_annotaitons that must exist e.g. = {'Presence': True}
    '''
    cui2pt2cnt = defaultdict(lambda: defaultdict(int))
    pt2stream = defaultdict(list)

    # Frequency for each each entity given a patient
    for doc in docs:
        for ent in docs[doc]['entities'].values():
            # Must match all meta meta_anns
            if not meta_requirements or all([ent['meta_anns'][name]['value'] == value for name, value in meta_requirements.items()]):
                cui = ent['cui']
                pt = doc2pt[doc]
                cui2pt2cnt[cui][pt] += 1

    for doc in docs:
        for ent in docs[doc]['entities'].values():
            if not meta_requirements or all([ent['meta_anns'][name]['value'] == value for name, value in meta_requirements.items()]):
                cui = ent['cui']
                pt = doc2pt[doc]
                if doc2time is not None:
                    timestamp = doc2time[doc]
                else:
                    timestamp = ent['document_timestamp']
                cnt = cui2pt2cnt[cui][pt]
                if ent[entity_type_column]: # This can be none in some cases
                    token_type = ent[entity_type_column][0]
                else:
                    raise Exception("Entity type must be set")

                pt2stream[pt].append((cui, cnt, timestamp, token_type))

    return dict(pt2stream) # Convert into a standard dict, we do not need the functionality anymore
